package w2phtml.rdf;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.TimeZone;
import java.util.Vector;
import java.text.SimpleDateFormat;
import org.apache.jena.ontology.OntClass;
import org.apache.jena.ontology.OntModel;
import org.apache.jena.rdf.model.ModelFactory;
import org.apache.jena.rdf.model.Resource;
import org.apache.jena.vocabulary.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import w2phtml.xhtml.XhtmlConfig;
import w2phtml.xhtml.XhtmlDocument;

import org.apache.jena.rdf.model.Property;


public class DocumentStructure {
	
	private static final String AFFILIATED_ORGANIZATION_POSTAL_CODE = "affiliated organization postal code";

	private static final String AFFILIATED_ORGANIZATION_ADDRESS = "affiliated organization address";

	private static final String AFFILIATED_ORGANIZATION_OFFICIAL_NAME = "affiliated organization official name";

	private static final String AFFILIATED_ORGANIZATION_NAME = "affiliated organization name";

	private static final String AUTHOR_INITIALS = "author initials";

	private static final String AUTHOR_EMAIL = "author email";

	private static final String AUTHOR_FAMILY = "author family";

	private static final Logger logger = LoggerFactory.getLogger(DocumentStructure.class);

	private static final String AUTHOR_GIVEN_NAME = "author given name";
	private static final String MODIFICATION_TIME = "modificationTime";
  private static final String HTML_EXCERPT_PROPERTY = "htmlExcerpt";
	private static final String TOCITEM = "TOCItem";
	private static final String TS = "https://litvinovg.pro/text_structures#";
	//private static final String PARSERNAME = "w2phtml"; 
	private static final String TEXT_EXCERPT = "textExcerpt";
	private static final String TOC_LEVEL = "TOCLevel";
	private static final String TABLE_OF_CONTENTS = "TOC";
	private static final String PARTICIPANT = "publicationParticipant";
	private static final String ORGANIZATION = "organization";
	
	private static final String PUBLICATION = "publication";
	private static final String BOOK = "book";
	private static final String JOURNAL = "journal";
	private static final String ENC_ARTICLE = "encArticle";
	private static final String ELENPH_ARTICLE = "elenphArticle";
	private static final String ELENPH_EXCERPT = "elenphExcerpt";
	
	private String docID = "DOC_ID";
	private String documentType = "elenphArticle";
	private String excerptType = "elenphExcerpt";
	
	private OntModel m;
	private OntModel textOntology;
	private OntClass excerptClass;
	private OntClass tocLevelClass;
	private OntClass itemClass;
	private OntClass documentClass;
	private OntClass tocClass;
	private OntClass organizationClass;
	private OntClass participantClass;


	private HashMap<String, Resource> tocLevels;
	private HashMap<String, DocumentPart> inputParts;
	private Metadata metadata;
  private Date currentTime;
	
	public DocumentStructure(Vector<XhtmlDocument> files,String fileName, XhtmlConfig config,Metadata metadata) {
		this.tocLevels = new HashMap<String, Resource>();
		this.inputParts = new HashMap<String, DocumentPart>(); 
		this.m = ModelFactory.createOntologyModel();
		this.metadata = metadata;
		this.textOntology = ModelFactory.createOntologyModel();
		this.tocLevelClass = textOntology.createClass(TS + TOC_LEVEL);
		this.tocClass = textOntology.createClass(TS + TABLE_OF_CONTENTS);
		this.organizationClass = textOntology.createClass(TS + ORGANIZATION);
		this.participantClass = textOntology.createClass(TS + PARTICIPANT);
		this.itemClass = textOntology.createClass(TS + TOCITEM);
		setDocID(fileName);
		this.currentTime = Calendar.getInstance().getTime();
		String publicationType = config.getRDFType();
		setPublicationType(publicationType);
		this.documentClass = textOntology.createClass(TS + documentType);
		this.excerptClass = textOntology.createClass(TS + excerptType);
		for(int i = 0 ; i< files.size();i++) {
			XhtmlDocument inputDoc = files.get(i);
			DocumentPart part = new DocumentPart(inputDoc, metadata);
			part.setOrder(Integer.toString(i));
			addPart(part);
		}
		addEmptyParts();
	}
	
	private void setPublicationType(String publicationType) {

		if (publicationType.equals(ELENPH_ARTICLE)) {
			this.documentType = ELENPH_ARTICLE;
			this.excerptType = ELENPH_EXCERPT;
		} else if (publicationType.equals(BOOK)){
			this.documentType = BOOK;
			this.excerptType = TEXT_EXCERPT;
		} else if (publicationType.equals(JOURNAL)){
			this.documentType = JOURNAL;
			this.excerptType = TEXT_EXCERPT;
		} else if (publicationType.equals(ENC_ARTICLE)){
			this.documentType = ENC_ARTICLE;
			this.excerptType = TEXT_EXCERPT;
		}
	}

	private void addEmptyParts() {
		Set<String> paths = inputParts.keySet();
		String[] array = new String[paths.size()];
		paths.toArray(array);
		for (int k = 0; k < array.length;k++) {
			String[] levels = array[k].split(" ");
			for (int i = 0; i < levels.length; i++) {
				if (levels[i].equals("0")) {
					String emptyPath = createEmptyPath(levels, i);
					if (!inputParts.containsKey(emptyPath)) {
						DocumentPart emptyPart = new DocumentPart(emptyPath, metadata);
						addPart(emptyPart);
					}
				}
			}	
		}
	}
	
	private String createEmptyPath(String[] levels, int i) {
		StringBuilder emptyPath = new StringBuilder();
		for (int j = 0; j<= i;j++) {
			if (j != 0) {
				emptyPath.append(" ");
			}
			emptyPath.append(levels[j]);
		}
		return emptyPath.toString();
	}
		
	public void printModel(String fileName) {
		File outFile = new File(fileName + ".rdf");
		FileWriter fw = null;
		try {
			outFile.createNewFile();
			fw = new FileWriter(outFile);
			m.write(fw,"RDF/XML-ABBREV");
			
		} catch (IOException e) {
			System.out.println("File couldn't be created");
			e.printStackTrace();
		} finally {
			try {
				fw.close();
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
	}
	public void printModel(OutputStream os) {
		m.write(os,"RDF/XML-ABBREV");
	}
	
	private void addPart(DocumentPart docExcerpt) {
		inputParts.put(docExcerpt.getPath(), docExcerpt);
	}
	
	private Resource createExcerpt(DocumentPart docExcerpt) {
		String excerptName = TS + excerptType + "/" + docID + docExcerpt.getSafePath();
		Resource excerpt = m.createIndividual(excerptName, excerptClass);
		addModificationTime(excerpt);
		if (!docExcerpt.getBody().isEmpty()) {
			Property htmlExcerpt = m.createProperty(TS + HTML_EXCERPT_PROPERTY);
			excerpt.addLiteral(htmlExcerpt, docExcerpt.getBody());	
		}
		return excerpt;
	}
	
	private void createTOCLevel(DocumentPart docPart) {
		String levelName = TS + TOC_LEVEL + "/" + docID + docPart.getSafePath();
		Resource level = m.createIndividual(levelName,tocLevelClass);
		level.addProperty( RDFS.label, docPart.getName());
		tocLevels.put(docPart.getPath(), level);
		attachExcerpt(docPart, level);		
	}	
	
	private void createDocument(DocumentPart docPart) {
		String documentURI = TS + documentType + "/" + docID ;
		Resource mainResource = m.createResource(documentURI, documentClass);
		mainResource.addProperty( RDFS.label, docPart.getName());
		
	  addModificationTime(mainResource);
		
		String tocURI = TS + TABLE_OF_CONTENTS + "/" + docID ;
		Resource toc = m.createResource(tocURI, tocClass);
		toc.addProperty( RDFS.label, docPart.getName());

		Property hasTOC = m.createProperty(TS + "hasTOC");
		mainResource.addProperty(hasTOC, toc);
				
		addMetadataProperties(mainResource, docPart);
		//addAuthor(mainResource, docPart);
		tocLevels.put(docPart.getPath(), toc);
		attachExcerpt(docPart, toc);	
	}

  private void addModificationTime(Resource mainResource) {
    Property property = m.createProperty(TS + MODIFICATION_TIME);
    SimpleDateFormat sd = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss");
    sd.setTimeZone(TimeZone.getTimeZone("GMT"));
    String todayAsString = sd.format(currentTime);
	  mainResource.addProperty( property, todayAsString);
  }
	
	private void addAuthor(Resource excerpt, DocumentPart docPart) {
		String order = docPart.getOrder();				
		ArrayList<Map<String, String>> sectionMeta = metadata.getSection(order);
		for (Map<String, String> map : sectionMeta) {
			Set<String> names = map.keySet();
			for (String givenNameKey : names) {
				if (givenNameKey.startsWith(AUTHOR_GIVEN_NAME)) {
					String postfix = givenNameKey.replaceFirst(AUTHOR_GIVEN_NAME, "");
					if (!postfix.matches("[a-zA-Z0-9]*")) {
						logger.error("Metadata author postfix is bad. Should contain only symbols a-zA-Z0-9, but "
						+ postfix + " found. Author givenName field is " + givenNameKey );
						System.exit(1);
					}
					String participantURI = TS + PARTICIPANT + "/" + docID + order + "author" + postfix;
					Resource participant = m.createResource(participantURI, participantClass);
					Property hasAuthor = m.createProperty(TS + "hasAuthor");
					excerpt.addProperty(hasAuthor, participant);
					String givenName = map.get(givenNameKey);
					Property givenNameProperty = m.createProperty(TS + "participantGivenName");
					participant.addProperty( givenNameProperty, givenName.trim());
					
					String family = map.get(AUTHOR_FAMILY + postfix);
					if (family != null) {
						Property familyProperty = m.createProperty(TS + "participantFamily");
						participant.addProperty( familyProperty, family.trim());
					}
					String labelName = givenNameKey.trim();
					if (family != null ) {
						labelName += " " + family;
					}
					participant.addProperty(RDFS.label, labelName);
					
					String email = map.get(AUTHOR_EMAIL + postfix);
					if (email != null) {
						Property emailProperty = m.createProperty(TS + "participantEmail");
						participant.addProperty( emailProperty, email.trim());
					}
					String authorInitials = map.get(AUTHOR_INITIALS + postfix);
					if (email != null) {
						Property initialsProperty = m.createProperty(TS + "participantInitials");
						participant.addProperty( initialsProperty, authorInitials.trim());
					}
					attachOrganization(participant, map, order, postfix);
				}
			}
		}
	}

	private void attachOrganization(Resource participant, Map<String, String> map, String order, String postfix) {
		String orgName = map.get(AFFILIATED_ORGANIZATION_NAME + postfix);
		if (orgName == null) {
			System.out.println(orgName + " not found");
			logger.warn("No " + AFFILIATED_ORGANIZATION_NAME + postfix + " found.");
			return;
		}
		String organizationUri = TS + ORGANIZATION + "/" + docID + order + "author" + postfix;
		Resource organization = m.createResource(organizationUri, organizationClass);
		Property affiliatedWith = m.createProperty(TS + "affiliatedWith");
		participant.addProperty(affiliatedWith, organization);
	
		organization.addProperty(RDFS.label, orgName);

		String officialName = map.get(AFFILIATED_ORGANIZATION_OFFICIAL_NAME + postfix);
		if (officialName != null) {
			Property officialNameProperty = m.createProperty(TS + "officialOrganizationName");
			organization.addProperty(officialNameProperty, officialName);
		}
		String orgAddress = map.get(AFFILIATED_ORGANIZATION_ADDRESS + postfix);
		if (orgAddress != null) {
			Property orgAddressProperty = m.createProperty(TS + "organizationAddress");
			organization.addProperty(orgAddressProperty, orgAddress);
		}
		String postalCode = map.get(AFFILIATED_ORGANIZATION_POSTAL_CODE + postfix);
		if (postalCode != null) {
			Property postalCodeProperty = m.createProperty(TS + "organizationPostalCode");
			organization.addProperty(postalCodeProperty, postalCode);
		}

	}

	private void addMetadataProperties(Resource resource, DocumentPart docPart) {
		
		String order = docPart.getOrder();				
		ArrayList<Map<String, String>> sectionMeta = metadata.getSection(order);
		for (Map<String, String> map : sectionMeta) {
			Set<String> names = map.keySet();
			for (String name : names) {
				String value = map.get(name);
				if (value != null && !value.isEmpty()) {
						addMetadata(resource,name,value.trim());
				}
			}
		}
	}
	private void addMetadata(Resource resource, String name, String value) {
		if (isNotBlacklisted(name)) {
			name = convertName(name);
			if (isDefinedInOntology(resource,name)) {
				if (name.equals("keywords")) {
					addKeywords(resource, name, value);
				} else {
					Property property = m.createProperty(TS + name);
					resource.addProperty( property, value);
				}
			}
		}
	}
	private void addKeywords(Resource resource,String name, String value) {
		String[] keywords = value.split(",");
		for (int i = 0; i < keywords.length; i++) {
			String keyword = keywords[i].trim();
			if (!keyword.isEmpty()) {
				Property property = m.createProperty(TS + name);
				resource.addProperty( property, keyword);
			}
		}
	}
	
	private boolean isDefinedInOntology(Resource resource, String name) {
		String nameSpace = resource.getNameSpace();
		if (nameSpace.contains(TS + excerptType)) {
			if (name.equals("bibliography") ||
					name.equals("keywords") ||
					name.equals("works")
					) {
				return true;
			}
		} else 
		if (nameSpace.contains(TS + documentType)) {
			if (name.equals("doi") ||
					name.equals("firstPublication") ||
					name.equals("year") ||
					name.equals("issue")
					) {
				return true;
			}
		}
		System.out.println("Skip " + name + " for " + resource.getNameSpace());
		return false;
	}
	private String convertName(String name) {
		if (name.equals("Affiliation")){
			return "affiliation";
		} else 
		if (name.equals("DOI")){
			return "doi";
		} else
		if (name.equals("1st-edition")){
			return "firstPublication";
		} else
		if (name.equals("year")){
			return "yearAndMonth";
		} else
		if (name.equals("year.short")){
				return "year";
		}else
		return name;
	}
	private boolean isNotBlacklisted(String name) {
		if (name.equals("Filename") ||
				name.equals("Section") ||
				name.equals("dc.Title") ||
				name.equals("subtitle") ||
				name.equals("dc.Identifier")
				) {
				return false;
		}
	
		return true;
	}
	private void createTOCItem(DocumentPart docPart) {
		String tocItemName = TS + TOCITEM + "/" + docID + docPart.getSafePath();
		Resource tocItem = m.createIndividual(tocItemName,itemClass);
		tocItem.addProperty( RDFS.label, docPart.getName());
		Property pointsTo = m.createProperty(TS + "pointsTo");
		Property itemNumber = m.createProperty(TS + "itemNumber");
		Property hasTOCItem = m.createProperty(TS + "hasTOCItem");	
		tocItem.addLiteral(itemNumber, docPart.getNumber());
		m.add(tocItem, pointsTo, tocLevels.get(docPart.getPath()));
		if (!docPart.getPath().isEmpty()) {
			Resource parent = tocLevels.get(docPart.getParentPath());
			m.add(parent, hasTOCItem, tocItem);
		}
	
	}
	
	private void attachExcerpt(DocumentPart docPart, Resource element) {
		if (docPart.isEmpty()) {
			return;
		}
		if (docPart.getBody().isEmpty()) {
			return;
		}
		Resource excerpt = createExcerpt(docPart);
		excerpt.addProperty( RDFS.label, docPart.getName());
		Property hasText = m.createProperty(TS + "hasText");
		element.addProperty(hasText, excerpt);
		if (!docPart.isMasterPart()) {
			addMetadataProperties(excerpt, docPart);
			addAuthor(excerpt, docPart);

		}
	}
	
	public void createTree() {
		createElements();
		createTOCItems();
	}
	
	private void createTOCItems() {
		Set<String> paths = inputParts.keySet();
		for (String path : paths) {
			DocumentPart part = inputParts.get(path);
			if (!part.getNumber().equals("")) {
				createTOCItem(part);	
			}
		}		
	}

	private void createElements() {
		Set<String> paths = inputParts.keySet();
		for (String path : paths) {
			DocumentPart part = inputParts.get(path);
			if (part.getPath().isEmpty()) {
				createDocument(part);
			} else {
				createTOCLevel(part);
			}
		}
	}
	
	private void setDocID(String fileName) {
		String identifier = fileName.replaceAll("\\s+", "_");
		identifier = transliterateToEn(identifier);
		identifier = identifier.replaceAll("[^a-zA-Z0-9_.-]", "");
		this.docID = identifier;
	}

	private String transliterateToEn(String fileName) {
    return fileName
    		.replaceAll("[аА]", "a")
    		.replaceAll("[бБ]", "b")
    		.replaceAll("[вВ]", "v")
    		.replaceAll("[гГ]", "g")
    		.replaceAll("[дД]", "d")
    		.replaceAll("[еЕ]", "e")
    		.replaceAll("[ёЁ]", "e")
    		.replaceAll("[жЖ]", "zh")
    		.replaceAll("[зЗ]", "z")
    		.replaceAll("[иИ]", "i")
    		.replaceAll("[йЙ]", "y")
    		.replaceAll("[кК]", "k")
    		.replaceAll("[лЛ]", "l")
    		.replaceAll("[мМ]", "m")
    		.replaceAll("[нН]", "n")
    		.replaceAll("[оО]", "o")
    		.replaceAll("[пП]", "p")
    		.replaceAll("[рР]", "r")
    		.replaceAll("[сС]", "s")
    		.replaceAll("[тТ]", "t")
    		.replaceAll("[уУ]", "u")
    		.replaceAll("[фФ]", "f")
    		.replaceAll("[хХ]", "kh")
    		.replaceAll("[цЦ]", "ts")
    		.replaceAll("[чЧ]", "ch")
    		.replaceAll("[шШ]", "sh")
    		.replaceAll("[щЩ]", "sch")
    		.replaceAll("[ъЪ]", "")
    		.replaceAll("[ыЫ]", "y")
    		.replaceAll("[ьЬ]", "")
    		.replaceAll("[эЭ]", "e")
    		.replaceAll("[юЮ]", "yu")
    		.replaceAll("[яЯ]", "ya");
	}

}
